import matplotlib
matplotlib.use('Agg')
from mpl_toolkits.mplot3d import Axes3D
import mpl_toolkits.mplot3d.art3d as art3d
import matplotlib.pyplot as plt
from numpy import linspace, meshgrid, sqrt, outer, sin, cos, ones, zeros, pi
from matplotlib.patches import Circle





def prepare_plot_space(x_lim,y_lim,z_lim):
    fig=plt.figure()
    ax = Axes3D(fig, azim=30, elev=30)
    ax.set_xlabel('x-axis')
    ax.set_xlim(x_lim)
    ax.set_ylabel('y-axis')
    ax.set_ylim(y_lim)
    ax.set_zlabel('z-axis')
    ax.set_zlim(z_lim)
    return(ax)


def plot_3D_cylinder(ax,radius, height, elevation=0, resolution=100, color='r', x_center = 0, y_center = 0):
    x =  linspace(x_center-radius, x_center+radius, resolution)
    z =  linspace(elevation, elevation+height, resolution)
    X, Z =  meshgrid(x, z)

    Y =  sqrt(radius**2 - (X - x_center)**2) + y_center # Pythagorean theorem

    ax.plot_surface(X, Y, Z, linewidth=0, color=color)
    ax.plot_surface(X, (2*y_center-Y), Z, linewidth=0, color=color)

    floor = Circle((x_center, y_center), radius, color=color)
    ax.add_patch(floor)
    art3d.pathpatch_2d_to_3d(floor, z=elevation, zdir="z")

    ceiling = Circle((x_center, y_center), radius, color=color)
    ax.add_patch(ceiling)
    art3d.pathpatch_2d_to_3d(ceiling, z=elevation+height, zdir="z")

def plot_sphere(ax,radius,elevation,color):

    u =  linspace(0, 2 *  pi, 100)
    v =  linspace(0,  pi, 100)
    x = radius *  outer( cos(u),  sin(v))
    y = radius *  outer( sin(u),  sin(v))+elevation
    z = radius *  outer( ones( size(u)),  cos(v))

    # Plot the surface
    ax.plot_surface(x, y, z, color=color)
    plt.show()
    #plt.close()




def plot_tree(ax,x,y,DBH,H,Crown_Length,Crown_Diameter):
    #plot stem
    plot_3D_cylinder(ax=ax,radius=DBH, height=H-Crown_Length, elevation=0, resolution=200, color="brown", x_center=x, y_center=y)
    #plot crown
    plot_3D_cylinder(ax=ax,radius=Crown_Diameter, height=Crown_Length, elevation=H-Crown_Length, resolution=100, color="green", x_center=x, y_center=y)



"""
plot_tree(ax=ax,x=30,y=20, DBH=1.5, H=15,Crown_Length=8,Crown_Diameter=5)
plot_tree(ax=ax,x=50,y=30, DBH=3, H=30,Crown_Length=12,Crown_Diameter=12)
plot_tree(ax=ax,x=80,y=80, DBH=1, H=10,Crown_Length=4,Crown_Diameter=5)
plt.show()
"""

def plot_NEE(s):
    #s=s*(-1)
    fig=plt.figure()
    ax1=plt.subplot2grid((1,1),(0,0))
    ax1.plot(range(len(s)),s,'-',label="NEE")
    ax1.plot([],[],linewidth=5,label="sink",color="b",alpha=0.5)
    ax1.plot([],[],linewidth=5,label="source",color="r",alpha=0.5)
    ax1.fill_between(range(len(s)),s,0,where=(s> zeros(len(s))),facecolor="b",alpha=0.5)
    ax1.fill_between(range(len(s)),s,0,where=(s< zeros(len(s))),facecolor="r",alpha=0.5)
    plt.title("Net Ecosystem Exchange")
    plt.xlabel("Time[yrs]")
    plt.ylabel("[to/ha/yr]")
    ax1.legend(loc='upper left', bbox_to_anchor=(1, 0.5))
    return fig
    #plt.savefig("NEE.png")
    #plt.close(fig)
    #plt.show()

def plot_emissions(dic):
    FastSoil=dic['soil_fast']
    SlowSoil=[f+s for f,s in zip(FastSoil,dic['soil_slow'])]
    DeadWood=[s+w for s,w in zip(SlowSoil,dic['dead_wood'])]
    LivingTrees=[d+l for d,l in zip(DeadWood,dic['living_trees'])]

    fig=plt.figure()
    ax1=plt.subplot2grid((1,1),(0,0))
    #ax1.plot(range(len(s)),s,'-',label="NEE")
    ax1.plot([],[],linewidth=5,label="Living Trees",color="green",alpha=0.5)
    ax1.plot([],[],linewidth=5,label="Dead Wood",color="orange",alpha=0.5)
    ax1.plot([],[],linewidth=5,label="Slow Soil",color="brown",alpha=0.5)
    ax1.plot([],[],linewidth=5,label="Fast Soil",color="yellow",alpha=0.5)

    length=len(dic['living_trees'])
    ax1.fill_between(range(length),LivingTrees,0,facecolor="green",alpha=0.8)
    ax1.fill_between(range(length),DeadWood,0,facecolor="orange",alpha=0.8)
    ax1.fill_between(range(length),SlowSoil,0,facecolor="brown",alpha=0.8)
    ax1.fill_between(range(length),FastSoil,0,facecolor="yellow",alpha=0.8)
    plt.title("Carbon Emmissions")
    plt.xlabel("Time[yrs]")
    plt.ylabel("Carbon [t]")
    ax1.legend(loc='upper center', bbox_to_anchor=(0.5, -0.05),
          fancybox=True, shadow=True, ncol=4)
    return fig

def plot_respiration(dic):
    Ft_col={"Cgpp":"forestgreen","Cr":"orange"}

    fig=plt.figure()
    ax1=plt.subplot2grid((1,1),(0,0))
    ax1.plot(dic["Cgpp"],linewidth=3,label="Cgpp",color=Ft_col["Cgpp"])
    ax1.plot(dic["Cr"],linewidth=3,label="Cr",color=Ft_col["Cr"])
    #plt.ylim(ymin=-10)
    plt.title("Respiration and Primary Production")
    plt.xlabel("Time[yrs]")
    plt.ylabel("Carbon")
    #plt.axhline(y=0)
    plt.legend(loc="upper right")
    return fig

def plot_populations(Pop):
    Ft_col={"FT1":"forestgreen","FT2":"orange","FT3":"saddlebrown","FT4":"mediumorchid","FT5":"crimson","FT6":"steelblue" }

    fig=plt.figure()
    ax1=plt.subplot2grid((1,1),(0,0))
    ax1.plot(Pop["FT1"],linewidth=3,label="FT1",color=Ft_col["FT1"])
    ax1.plot(Pop["FT2"],linewidth=3,label="FT2",color=Ft_col["FT2"])
    ax1.plot(Pop["FT3"],linewidth=3,label="FT3",color=Ft_col["FT3"])
    ax1.plot(Pop["FT4"],linewidth=3,label="FT4",color=Ft_col["FT4"])
    ax1.plot(Pop["FT5"],linewidth=3,label="FT5",color=Ft_col["FT5"])
    ax1.plot(Pop["FT6"],linewidth=3,label="FT6",color=Ft_col["FT6"])
    #plt.ylim(ymin=-10)
    plt.title("Functional Type Population")
    plt.xlabel("Time[yrs]")
    plt.ylabel("Population Size")
    #plt.axhline(y=0)
    ax1.legend(loc='upper center', bbox_to_anchor=(0.5, -0.05),
          fancybox=True, shadow=True, ncol=6)
    return fig
    #plt.show()
    #plt.close(fig)

def plot_stocks(Stocks):
    Soil=[s+f for s,f in zip(Stocks["Sslow"],Stocks["Sfast"])]
    DeadWood=[s+w for s,w in zip(Soil,Stocks["Dwood"])]
    AGB=[d+a for d,a in zip(DeadWood,Stocks["AGB"])]
    fig=plt.figure()
    ax1=plt.subplot2grid((1,1),(0,0))
    #ax1.plot(range(len(s)),s,'-',label="NEE")
    ax1.plot([],[],linewidth=5,label="AGB",color="green",alpha=0.5)
    ax1.plot([],[],linewidth=5,label="Dead Wood",color="orange",alpha=0.5)
    ax1.plot([],[],linewidth=5,label="Soil",color="brown",alpha=0.5)
    ax1.fill_between(range(len(AGB)),AGB,0,facecolor="green",alpha=0.5)
    ax1.fill_between(range(len(DeadWood)),DeadWood,0,facecolor="orange",alpha=0.5)
    ax1.fill_between(range(len(Soil)),Soil,0,facecolor="brown",alpha=0.5)
    plt.title("Carbon Stocks")
    plt.xlabel("Time[yrs]")
    plt.ylabel("Carbon [t]")
    ax1.legend(loc='upper center', bbox_to_anchor=(0.5, -0.05),
          fancybox=True, shadow=True, ncol=3)
    return fig
    #plt.show()
    #plt.close(fig)






def plot_ground_light(topology):
    plt.imshow(topology.surface[topology.dim_names["GroundLight"]], cmap=plt.get_cmap("gray"),vmin=0,vmax=860)
    plt.colorbar()
    plt.show()

def plot_tree_pos(Tree,output_dir,length,xlim,ylim,step):
    x=[]
    y=[]
    s=[]
    tp=[]
    Ft_col={"FT1":"forestgreen","FT2":"orange","FT3":"saddlebrown","FT4":"mediumorchid","FT5":"crimson","FT6":"steelblue" }

    for t in Tree.Instances.values():
        x.append(t.position[0])
        y.append(t.position[1])
        s.append(t.CA)
        tp.append(Ft_col[t.Ftype])

    fig=plt.figure()
    ax = fig.add_subplot(111, aspect='equal')
    #ax.autoscale(False)
    ax.set_xlim(-1,xlim+1)
    ax.set_ylim(-1,ylim+1)
    width, height = fig.canvas.get_width_height()
    p=length/width

    areas=[a/p for a in s]


    ax.scatter(x,y,s=areas,c=tp,alpha=0.5)
    plt.title("Simulation Year:"+str(step))
    plt.savefig(output_dir+"step: "+str(step)+".png")
    plt.close(fig)

def add_time_guide(figname,length,plot_func,**kwargs):
    """
    Plot the result of [plot_func] [length] times, with a vetical line moving along the x (time) axis.

    plot_func() must return a pylab fig.
    """
    for i in range(length):
        fig=plot_func(**kwargs)
        plt.axvline(x=i)
        plt.savefig(figname+str(i)+".png")
        plt.close(fig)



from py_ibm import *
#from plot_trees import plot_tree_pos
import pdb
from tables import *
import json
import numpy as np

class Tree(Agent):
    """ Represent individual trees.

    Args:
        position (tuple): (x,y) coordinates of where tree will be stablished.
        world (world object): world where tree will live.
        dbh (float): Diameter at Breast Height (in cm).
        h0 (float): Height-stem diameter relationship parameter.
        h1 (float): Height-stem diameter relationship parameter.
        cl0 (float): Crown Length-Height relationship parameter.
        cd0 (float): Crown Diameter-Stem Diameter relationship parameter.
        cd1 (float): Crown Diameter-Stem Diameter relationship parameter.
        cd2 (float): Crown Diameter-Stem Diameter relationship parameter.
        rho (float): wood density in t of Organic Dry Matter/cubic meter.
        sigma (float): ratio of total aboveground biomass to stem biomass.
        f0 (float): form factor-stem diameter relationship parameter.
        f1 (float): form factor-stem diameter relationship parameter.
        l0 (float): Leaf Area Index (LAI)-Stem relationship parameter.
        l1 (float): Leaf Area Index (LAI)-Stem relationship parameter.
        m (float): Light trasnmission coefficient.
        alpha (float): quantum eciency; initial slope of the type specic light response curve.
        pmax (float): maximum leaf gross photosynthetic rate.
        rg (float): fraction of gross primary production available for growth that is attributed to growth respiration.
        deltaDmax (float): maximum diameter increment.
        DdeltaDmax (float): % of Dmax that reaches deltaDmax.
        age (int): age (years).
        Ftype (str): functional type.
        Mb (float): backgroung mortality probability.
        Dfall (float): mimimum dbh for a tree to fall.
        pfall (float): probabbility that a tree with dbh>Dfall will fall.
        m_max (float): maximum size-dependent mortality for small trees.
        Dmort (float): DBH up to which mortality is increased (for small trees).

    Attributes:

        ID (int):
        Instances (dict):
        DeadTrees (dict):

        patch (tuple): (x,y) patch in which tree is located.
        f (float):
        k (float):
        lday (float):
        phi_act (float):

        H (float): Height (m).
        CL (float): Crown length (m).
        CD (float): Crown diameter (m)
        CA (float): Crown area (sq. m)
        AGB (float): Above ground biomass (tODM)
        LAI (float): Leaf area index.
        lmax (float):
        lmin (float):
        L_mean (float):
        Li (float):
        Mc (float):
        basic_m (float):
        Mb (float):
        Md (float):
        GPP (float): Gross primary production in this time step.
        Rm (float): Respiration maintenance in this timestep.
        position (tuple): (x,y) coordinates of where tree is located.
        world (world object): world where tree lives.
        dbh (float): Diameter at Breast Height (in cm).
        h0 (float): Height-stem diameter relationship parameter.
        h1 (float): Height-stem diameter relationship parameter.
        cl0 (float): Crown Length-Height relationship parameter.
        cd0 (float): Crown Diameter-Stem Diameter relationship parameter.
        cd1 (float): Crown Diameter-Stem Diameter relationship parameter.
        cd2 (float): Crown Diameter-Stem Diameter relationship parameter.
        rho (float): wood density in t of Organic Dry Matter/cubic meter.
        sigma (float): ratio of total aboveground biomass to stem biomass.
        f0 (float): form factor-stem diameter relationship parameter.
        f1 (float): form factor-stem diameter relationship parameter.
        l0 (float): Leaf Area Index (LAI)-Stem relationship parameter.
        l1 (float): Leaf Area Index (LAI)-Stem relationship parameter.
        m (float): Light trasnmission coefficient.
        alpha (float): quantum eciency; initial slope of the type-specic light response curve.
        pmax (float): maximum leaf gross photosynthetic rate.
        rg (float): fraction of gross primary production available for growth that is attributed to growth respiration.
        deltaDmax (float): maximum diameter increment.
        DdeltaDmax (float): % of Dmax that reaches deltaDmax.
        age (int): age (years).
        Ftype (str): functional type.
        Mb (float): backgroung mortality probability.
        max_height (float): maximum height this tree can reach.
        Dfall (float): mimimum dbh for a tree to fall.
        pfall (float): probabbility that a tree with dbh>Dfall will fall.
        m_max (float): maximum size-dependent mortality for small trees.
        Dmort (float): DBH up to which mortality is increased (for small trees).

    """


    ID=0
    Instances={}
    DeadTrees={"FT1":0,"FT2":0,"FT3":0,"FT4":0,"FT5":0,"FT6":0}
    DERIVED_TREES={}

    @classmethod
    def TreeFactory(cls,new_cls_name,new_parameters):

        def new_init(self,position,world,dbh,age=0,id=None, **kwargs):

            self.DBH=dbh
            self.age=age

            super().__init__(position,world,id=id)

            self.patch=(int(np.floor(self.position[0])),int(np.floor(self.position[1])))
            self.h0=new_parameters['h0']
            self.h1=new_parameters['h1']
            self.cl0=new_parameters['cl0']
            self.cd0=new_parameters['cd0']
            self.cd1=new_parameters['cd1']
            self.cd2=new_parameters['cd2']
            self.rho=new_parameters['rho']
            self.sigma=new_parameters['sigma']
            self.f0=new_parameters['f0']
            self.f1=new_parameters['f1']
            self.f=self.calculate_f()
            self.l0=new_parameters['l0']
            self.l1=new_parameters['l1']
            self.k=0.7
            self.lday=12
            self.phi_act=30
            self.m=new_parameters['m']
            self.alpha=new_parameters['alpha']
            self.pmax=new_parameters['pmax']
            self.Dmax=new_parameters['max_dbh']/100.
            self.deltaDmax=new_parameters['deltaDmax']
            self.DdeltaDmax=new_parameters['DdeltaDmax']
            self.Dmort=new_parameters['Dmort']
            self.Dfall=new_parameters['Dfall']
            self.pfall=new_parameters['pfall']
            self.alpha0=self.calculate_alpha0()
            self.alpha1=self.calculate_alpha1()
            self.rg=new_parameters['rg']
            self.Ftype=new_cls_name
            self.m_max=new_parameters['m_max']

            self.H=0
            self.CL=0
            self.CD=0
            self.CA=0
            self.AGB=0
            self.LAI=0
            self.lmax=0
            self.lmin=0
            self.L_mean=0
            self.Li=0
            self.Mc=0
            self.basic_m=new_parameters['Mb']
            self.Mb=0
            self.Md=0
            self.GPP=0 #Gpp in this time step
            self.Rm=0 #Rm in this timestep
            self.available_fruits=0
            #self.update()
            self.world.topology.trees_per_patch[self.patch].append(self.id)

            if id == None:
                #self.id=Tree.ID
                self.Indices.append(self.id)
                Tree.Instances[self.id]=self
                Tree.IncrementID()
            else:
                self.id=id
                self.Indices.append(self.id)
                Tree.Instances[self.id]=self






        new_attr_met={'Nseed':new_parameters['Nseed'],
                    'Indices':[],
                    'Ftype':new_cls_name,
                    'Iseed':new_parameters['Iseed'],
                    'h0':new_parameters['h0'],
                    'h1':new_parameters['h1'],
                    '__init__':new_init}


        new_cls=type(new_cls_name, (cls,), new_attr_met)

        cls.DERIVED_TREES[new_cls_name]=new_cls

        return new_cls

    @classmethod
    def FTPopulation(cls,ft):
        pass

    @classmethod
    def DisperseSeeds(cls):
        for t in cls.Instances.values():
            t.disperse_seeds(a=0.1,fraction=0.5)

    @classmethod
    def TreesAboveDBH(self,dbh):
        """Calculate the total number of trees with DBH above the speciefied threshold.

        Returns:
            (int): Number of trees.
        """

        trees=[k for k in self.Instances.keys() if self.Instances.get(k).DBH>=dbh]
        return len(trees)

    @classmethod
    def Total_AGB(self):
        """ Calculate the total Above Ground Biomass for all living trees.

        Returns:
            (float): total AGB in t of Organic Dry matter (tODM).
        """

        Tagb=0
        trees=[k for k in self.Instances.keys()]
        for i in trees:
            t=Tree.Instances.get(i)
            Tagb+=t.AGB
        return Tagb

    @classmethod
    def Cgpp(self):
        """ Calculate the total Carbon absorbed as Gross Primary Production.

        Returns:
            (float): Total Carbon absorbed.
        """


        Cgpp=0
        trees=[k for k in self.Instances.keys()]
        for i in trees:
            t=Tree.Instances.get(i)
            Cgpp+=t.GPP
        return 0.44*Cgpp

    @classmethod
    def Cr(self):
        """ Calculate the total Carbon released through respiration.

        Returns:
            (float): Total Carbon released by all living trees.
        """
        Cr=0
        trees=[k for k in self.Instances.keys()]
        for i in trees:
            t=Tree.Instances.get(i)
            #Cr+=(t.Rm+t.rg*(t.GPP-t.Rm))
            Cr+=(t.Rm)
        return 0.44*Cr

    @classmethod
    def Mortality(self):
        """
        Kill trees by calling their stochastic_M() method.
        """
        trees=[k for k in Tree.Instances.keys()]
        for t in trees:
             if Tree.Instances.get(t).stochastic_M(): Tree.Instances.get(t).remove_me()


    @classmethod
    def BackgroundMortality(self):
        """
        Kill trees by calling their stochastic_Bm() method.
        """
        trees=[k for k in self.Instances.keys()]
        for t in trees:
             if Tree.Instances.get(t).stochastic_Bm(): Tree.Instances.get(t).remove_me()

    @classmethod
    def CrowdingMortality(self):
         """
         Kill trees by calling their stochastic_Cm() method.
         """
         trees=[k for k in self.Instances.keys()]
         for t in trees:
              if Tree.Instances.get(t).stochastic_Cm(): Tree.Instances.get(t).remove_me()

    @classmethod
    def DamageMortality(self):
          """
          Kill trees by calling their stochastic_Dm() method.
          """
          trees=[k for k in self.Instances.keys()]
          for t in trees:
               if Tree.Instances.get(t).stochastic_Dm(): Tree.Instances.get(t).remove_me()


    @classmethod
    def ActiveAgents(self):
        return [ag.id for ag in self.Instances.values() if ag.active == True]

    @classmethod
    def TreesPerPatch(self):
        """Identify which trees are in each patch.

        Returns:
            (dict): keys are patch coordinates '(x,y)' and values are lists of trees ids in that patch.

        """
        patches={}
        for t in self.Instances.values():
            if patches.get(t.patch)==None:
                patches[t.patch]=[t.id]
            else:
                patches[t.patch].append(t.id)
        return patches

    @classmethod
    def MaxHeight(self):
        """ Finds the height of the tallest tree in the current time step.

        Returns:
            (float): height of the tallest tree.

        """
        return max([t.H for t in self.Instances.values()],default=0)


    def __repr__(self):
        return "Tree type: {0}".format(self.Ftype)

    def __init__(self,position,world,dbh,wood_density,h0,h1,cl0,cd0,cd1,cd2,rho,sigma,f0,f1,l0,l1,m,alpha,pmax,max_agb,rg,deltaDmax,DdeltaDmax,age,Ftype,Mb,Dfall=45, pfall=0.3,m_max=0.12,Dmort=10,max_dbh=None,id=None):

        super().__init__(position,world,id=id)
        #self.AddInstance(self.id)
        #Tree.Instances[Ftype+' '+str(self.id)]=self

        self.patch=(int(np.floor(self.position[0])),int(np.floor(self.position[1])))
        self.DBH=dbh
        self.age=age
        self.Ftype=Ftype
        self.wood_density=wood_density
        #type specific parameter used to calculate tree height
        self.h0=h0
        self.h1=h1
        #type specific parameter used to calculate crown length
        self.cl0=cl0
        #type specific parameter used to calculate crown diameter
        self.cd0=cd0
        self.cd1=cd1
        self.cd2=cd2
        #type specific parameter used to calculate aboveground biomass
        self.rho=rho
        self.sigma=sigma
        self.f0=f0
        self.f1=f1
        self.f=self.calculate_f()
        #type specific parameter used to calculate the leaf area index
        self.l0=l0
        self.l1=l1
        self.k=self.world.topology.k
        self.lday=self.world.topology.lday
        self.phi_act=self.world.topology.phi_act
        self.m=m
        self.alpha=alpha
        self.pmax=pmax
        self.m_max=m_max
        self.Dmort=Dmort
        self.Dfall=Dfall
        self.pfall=pfall
        self.Dmax=max_dbh/100.
        self.deltaDmax=deltaDmax
        self.DdeltaDmax=DdeltaDmax
        self.alpha0=self.calculate_alpha0()
        self.alpha1=self.calculate_alpha1()
        self.rg=rg

        self.H=0
        self.CL=0
        self.CD=0
        self.CA=0
        self.AGB=0
        self.LAI=0
        self.lmax=0
        self.lmin=0
        self.L_mean=0
        self.Li=0
        self.Mc=0
        self.basic_m=Mb
        self.available_fruits=0
        self.Mb=0
        self.Md=0
        self.GPP=0 #Gpp in this time step
        self.Rm=0 #Rm in this timestep
        #self.update()
        self.world.topology.trees_per_patch[self.patch].append(self.id)


    def update(self):
        """ Updates geometry related parameters and mortality probabilities.
        """
        self.H=self.H_from_DBH()
        self.CL=self.CL_from_H()
        self.CD=self.CD_from_DBH()
        self.CA=self.CA_from_CD()
        if self.age==0:
            self.AGB=self.AGB_from_DBH(D=self.DBH/100)
        self.LAI=self.LAI_from_DBH()

        self.lmax=self.H/self.world.topology.delta_h #eq.31
        self.lmin=(self.H-self.CL)/self.world.topology.delta_h #eq.32
        self.L_mean=(self.LAI*self.CA)/(self.lmax-self.lmin if self.lmax-self.lmin!=0 else 0.5 )#eq.34
        self.Li=self.Lind()
        self.Mc=self.calculate_Rc()
        self.Mb=max(0,self.basic_m+self.calculate_Ms())
        self.Md=0
        self.available_fruits=self.fruits_from_DBH()


        attributes={"DBH":self.DBH,
                "H":self.H,
                "CL":self.CL,
                "CD":self.CD,
                "CA":self.CA,
                "AGB":self.AGB,
                "LAI":self.LAI,
                "lmax":self.lmax,
                "lmin":self.lmin,
                "L_mean":self.L_mean,
                "Li":self.Li,
                "Mc":self.Mc,
                "Mb":self.Mb,
                "age":self.age,
                "available_fruits":self.available_fruits,}

        return (attributes)

    def import_attributes(self,attributes):
        self.DBH=attributes["DBH"]
        self.H=attributes["H"]
        self.CL=attributes["CL"]
        self.CD=attributes["CD"]
        self.CA=attributes["CA"]
        self.AGB=attributes["AGB"]
        self.LAI=attributes["LAI"]
        self.lmax=attributes["lmax"]
        self.lmin=attributes["lmin"]
        self.L_mean=attributes["L_mean"]
        self.Li=attributes["Li"]
        self.Mc=attributes["Mc"]
        self.Mb=attributes["Mb"]
        self.age=attributes["age"]
        self.Md=0
        self.available_fruits=attributes["available_fruits"]
        #self.age+=1



    def fruits_from_DBH(self):
        return 30

    def H_from_DBH(self):
        """Calculate tree height calculated according to eq.1 in SI-Fisher et al.(2015).

        Returns:
            (float): H.
        """
        return self.h0*self.DBH**self.h1

    def CL_from_H(self):
        """Calculate tree crown_length (CL) calculated according to eq.2 in SI-Fisher et al.(2015)

        Returns:
            (float): CL.
        """
        return self.cl0*self.H

    def CD_from_DBH(self):
        """Calculate tree crown diameter (CD) calculated according to eq.3 in SI-Fisher et al.(2015).

        Returns:
            (float): CD.
        """
        return (self.cd0*self.DBH**self.cd1)-self.cd2

    def CA_from_CD(self):
        """Calculate tree crown area (CA) calculated according to eq.4 in SI-Fisher et al.(2015)

        Returns:
            (float): CA.
        """
        return (np.pi/4)*(self.CD**2)

    def calculate_volume(self):
        """ Calculate the volume of the stem based on DBH.

        Returns:
            (float): volume.
        """

        return self.H*(np.pi*((self.DBH/100)/2)**2)


    def AGB_from_DBH(self,D):
        """ Calculate tree aboveground biomass (AGB) calculated according to eq.5 in SI-Fisher et al.(2015)

        Returns:
            (float): AGB.
        """

        return (np.pi/4)*(D**2)*self.H*self.f*(self.rho/self.sigma)

    def calculate_f(self):
        """Calculate the form factor (f) calculated according to eq.6 in SI-Fisher et al.(2015)

        Returns:
            (float): f.

        """

        return self.f0*self.DBH**self.f1

    def LAI_from_DBH(self):
        """Calculate the leaf area index (LAI) calculated according to eq.7 in SI-Fisher et al.(2015)

        Returns:
            (float): LAI.
        """

        return self.l0*self.DBH**self.l1

    def decrement_fruit_stock(self,n=1):
        self.available_fruits-=n

    def disperse_seeds(self,fraction=0.5,a=0.3):
        n_seeds=int(round(self.available_fruits*fraction))
        self.decrement_fruit_stock(n=n_seeds)
        seed_distances=np.random.power(a=a,size=n_seeds)
        for d in seed_distances:
            seed_position=self.random_seed_position(distance=d)
            self.deposit_seed(seed_position)

    def random_seed_position(self, distance):
        x0,y0=self.position
        angle=np.random.randint(360)
        new_position=((x0+cos(radians(angle))*distance,y0+sin(radians(angle))*distance))
        verified_position=self.world.topology.in_bounds(new_position)
        return verified_position

    def deposit_seed(self,seed_position):
        self.world.topology.seedbank[self.Ftype].append(seed_position)
        self.world.add_seed_entry(tree_id=self.id,initial_x=self.position[0],
            initial_y=self.position[1],final_x=seed_position[0],final_y=seed_position[1],dispersal_type="non-zoochoric")





    def Lind(self):
        """Calculate the incoming radiation on top of lmax layer the tree is reaching according to eq.36 in SI-Fisher et al.(2015).

        Returns:
            (float): Incoming radiation on top of this tree.
        """

        #crown_layers=(np.ceil(self.lmax)-np.floor(self.lmin))/self.world.topology.delta_h
        patch_LI=self.world.topology.LAIs[self.patch]
        top_layer=int(np.ceil(self.lmax))#/self.world.topology.delta_h)
        patch_LI=patch_LI[top_layer:]

        #self.patch_LI=sum(patch_LI)

        return self.world.topology.I0*np.exp(-self.k*sum(patch_LI))



    def Pind(self,Li):
        """ Calculate the interim gross photosynthetic rate of one tree per year, according to eq.40 in SI-Fisher et al.(2015)

        Returns:
            (float) Gross photosynthetic rate.
        """

        pmk=self.pmax/self.k
        aki=self.alpha*self.k*Li
        pm=self.pmax*(1-self.m)

        pind=pmk*np.log((aki+pm)/(aki*np.exp(-self.k*self.LAI)+pm))

        #Tons of organic dry meter per year (todm.y-1)
        return pind*self.CA*60*60*self.lday*self.phi_act*np.power(10.0,-12)*0.63*44


    def calculate_rm(self):
            """Calculate the maintenance respiration rate(rm) according to eq.46 in SI-Fisher et al.(2015).

            Returns:
                (float): rm.
            """

            max_Li=self.world.topology.I0

            rm=(1/self.AGB)*(self.Pind(max_Li)-(max(0,(self.AGB_from_DBH(D=(self.DBH/100)+self.growth(D=self.DBH/100.))-self.AGB))/(1-self.rg)))

            return rm

    def growth(self,D):
        """Calculate the yearly diameter growth according to eq.48 in SI-Fisher et al.(2015). Assumes full availabilty of resources

        Returns:
            (float):diameter growth.
        """

        return (self.alpha0*D*(1-(D/self.Dmax)))*np.exp(-self.alpha1*D)

    def calculate_alpha0(self):
        """Calculate the alpha0 growth paramenter, according equations described in sctions F-5 (page 24) of SI-Fisher et al.(2015).

        Returns:
            (float): alpha 0.

        """


        exp=np.exp((self.Dmax-2*(self.DdeltaDmax*self.Dmax))/(self.Dmax-(self.DdeltaDmax*self.Dmax)))
        alpha0=exp*self.Dmax*self.deltaDmax
        alpha0=alpha0/((self.Dmax-(self.DdeltaDmax*self.Dmax))*(self.DdeltaDmax*self.Dmax))

        return alpha0

    def calculate_alpha1(self):
        """Calculate alpha1 growth paramenter, according equations described in sctions F-5 (page 24) of SI-Fisher et al.(2015).

        Returns:
            (float): alpha1.
        """

        alpha1=self.Dmax-2*(self.DdeltaDmax*self.Dmax)
        alpha1=alpha1/(self.Dmax*(self.DdeltaDmax*self.Dmax)-(self.DdeltaDmax*self.Dmax)**2)

        return alpha1

    def Biomass_gain(self):
        """ Calculate Aboveground Biomass (AGB) gain according to eq.43 in SI-Fisher et al.(2015).

        Returns:
            (float): AGB.
        """

        self.GPP=self.Pind(self.Li)
        self.Rm=0.47*self.GPP
        #self.Rm=self.calculate_rm()*self.AGB
        #self.Rm=self.r0*self.AGB+self.r1*self.AGB**2+self.r2*self.AGB**3
        gain=(1-self.rg)*(self.GPP-self.Rm)
        gain=self.GPP-self.Rm
        self.AGB+=gain
        return gain

    def DBH_from_Biomass(self,B):
        """ Calculate diameter at breast height from above ground biomass.

        Returns:
            (float): the DBH calculated from AGB.
        """


        p=(np.pi/4)*self.H*self.f*(self.rho/self.sigma)
        #if p>0 and B>0:
        dbh=np.sqrt(B/p)
        if isnan(dbh) or dbh<=0: return 0 #if Biomass is negative, new DBH will be 0
        return dbh

    def increase_DBH(self):
        """Increase tree DBH according to gain in biomass.

        Note: If DBH decreases to zero, remove tree.

        Returns:
            None.
        """
        gain=self.Biomass_gain()
        if round(self.AGB,3) <=0 or gain<0:
            self.remove_me()
        else:
            #new_B=self.AGB+self.Biomass_gain()
            self.DBH=self.DBH_from_Biomass(self.AGB)*100
            if round(self.DBH,3)<=2: self.remove_me()
        #if self.age>150: self.remove_me()

        #new_B=self.AGB+self.Biomass_gain()
        #self.DBH=self.DBH_from_Biomass(new_B)*100
        #if self.DBH<=0: self.remove_me()

    def calculate_Rc(self):
        patch_CCA=self.world.topology.CCAs[self.patch]
        max_CCA=max(patch_CCA[int(np.ceil(self.lmin)):int(np.ceil(self.lmax))],default=self.CA)

        if max_CCA==0: return 1.0
        rc=1/max_CCA
        if rc>=1.0:
            return 0
        else:
            return (1-rc)

    def calculate_Ms(self):
        """Calculate the size-dependent mortality for small trees according to model description in Ruger et al. 2007. (SI, page 4).

        Returns:
            (float): Ms.
        """
        Ms=0


        if self.DBH <self.Dmort:
            Ms=self.m_max-self.m_max*(self.DBH/self.Dmort)
        #elif self.DBH>=self.max_dbh:
        #    Ms=self.m_max

        return Ms


    def stochastic_M(self):
        """Stochastic mortality.

        Returns:
            (bools): True if tree should be removed.
        """
        #if self.DBH>=self.max_dbh:
        #    return True #np.random.random()<= self.m_max
        w_Mb=0
        w_Mc=1
        w_Md=1


        M=((w_Mc*self.Mc+w_Mb*self.Mb+w_Md*self.Md)/(w_Mc+w_Mb+w_Md))#
        return np.random.random()<=M#*self.world.mortality_factor

    def stochastic_Cm(self):
        """Stochastic crownding mortality.

        Returns:
            (bool): True if tree dies from crowding.
        """
        return np.random.random()<=min(self.Mc,1.0)

    def stochastic_Bm(self):
        """Stochastic Background mortality.

        Returns:
            (bool): True if tree dies from background mortality.
        """
        return np.random.random()<=min(self.Mb,1.0)

    def stochastic_Dm(self):
        """Stochastic damage mortality.

        Returns:
            (bool): True if tree dies from damage.
        """
        return np.random.random()<=min(self.Md,1.0)

    def log_me(self):
        """Cut down this tree.

            Randomly determines where tree is going to fall and inflict damage in the the trees hit by the crown. Only trees with DBH< 50 cm are damaged.

            The carbon of this trees is not added to any Stock.
        """


        subclass="Tree_"+self.Ftype
        globals()[subclass].Indices.remove(self.id)
        self.world.topology.trees_per_patch[self.patch].remove(self.id)
        #self.world.Smort+=self.AGB*0.44

        topology=self.world.topology

        dir=np.random.randint(1,360)
        x=self.position[0]+self.H*np.sin(np.deg2rad(2*np.pi*(dir/360)))
        y=self.position[1]+self.H*np.cos(np.deg2rad(2*np.pi*(dir/360)))
        md=min(self.CA/topology.patch_area,1)

        target_patch=np.floor((x,y))
        target_patch=(int(target_patch[0]),int(target_patch[1]))
        target_patch=topology.in_bounds(target_patch)



        neighborhood=topology.hood(cell=target_patch,remove_center=False)
        trees_in_neighborhood=topology.trees_in_patches(neighborhood)
        trees_in_circle=[id for id,p in trees_in_neighborhood if topology.point_within_circle(x,y,(self.CD/2),p[0],p[1],topology.scaled_distance,scale=topology.patch_area**0.5)]

        for t in trees_in_circle:
            if Tree.Instances.get(t).DBH <=50:
                 Tree.Instances.get(t).Md=md



        del Tree.Instances[self.id]


    def remove_me(self):
        """
        Kill this tree.
        """
        # subclass="Tree_"+self.Ftype
        # globals()[subclass].Indices.remove(self.id)
        Tree.DERIVED_TREES[self.Ftype].Indices.remove(self.id)
        if self.DBH>self.Dfall and np.random.random()<=self.pfall:
            self.fall()
        self.world.topology.trees_per_patch[self.patch].remove(self.id)
        self.world.Smort+=max(0,self.AGB)*0.44
        Tree.DeadTrees[self.Ftype]+=1
        del Tree.Instances[self.id]

    def fall(self):
        """
        Inflict damage on other trees affected by the falling one.
        """
        topology=self.world.topology

        dir=np.random.randint(1,360)
        x=self.position[0]+self.H*np.sin(np.deg2rad(2*np.pi*(dir/360)))
        y=self.position[1]+self.H*np.cos(np.deg2rad(2*np.pi*(dir/360)))
        md=min(self.CA/topology.patch_area,1)

        #Damage caused on trees near the falling trees, due to lianas

        neighborhood=topology.hood(cell=self.patch,remove_center=False)
        trees_in_neighborhood=topology.trees_in_patches(neighborhood)
        trees_in_circle=[id for id,p in trees_in_neighborhood if topology.point_within_circle(self.position[0],self.position[1],(self.CD/2),p[0],p[1],topology.scaled_distance,scale=topology.patch_area**0.5)]

        for t in trees_in_circle:
            if Tree.Instances.get(t).H <=(self.H+1):
                 Tree.Instances.get(t).Md=md*(topology.surface[topology.dim_names["LianaCoverage"]][self.patch])



        target_patch=np.floor((x,y))
        target_patch=(int(target_patch[0]),int(target_patch[1]))
        target_patch=topology.in_bounds(target_patch)



        neighborhood=topology.hood(cell=target_patch,remove_center=False)
        trees_in_neighborhood=topology.trees_in_patches(neighborhood)
        trees_in_circle=[id for id,p in trees_in_neighborhood if topology.point_within_circle(x,y,(self.CD/2),p[0],p[1],topology.scaled_distance,scale=topology.patch_area**0.5)]

        for t in trees_in_circle:
            if Tree.Instances.get(t).H <=(self.H+1):
                 Tree.Instances.get(t).Md=md



from py_ibm import *
#from plot_trees import plot_tree_pos
from trees_ibm.tree_agent import Tree
import pdb
from tables import *
import json
import numpy as np


class Tree_World(World):
    """ Provide observer methods and control the schedule and execution of simulations.


    Args:
        topology (Tree_Grid instance): grid object on which agents will live.
        AET (int): actual evapotranspiration in the previous year in mm.
        Assumed to be constant every year.

    Attributes:
        Smort (float): Carbon of all trees that died in one year [tC]
        Sdead (float): Dead wood carbon stock [tC]
        tsdead (float): Anual decomposition rate
        tsdead_A (float): Proportion of decomposed dead wood that is released to the atmosphere.
        Sslow (float): Slow decomposition portion of soil stock [tC]
        tsdead_Sslow (float): Proportion of decomposed dead wood that is tranferred to the slow decomposition Soil stock
        self.tsslow_A (float):Proportion of slow decomposition soil stock that is released to the atmosphere.
        Sfast (float):Fast decomposition portion of soil stock [tC]
        tsfast_A (float):Proportion of fast decomposition soil stock that is released to the atmosphere.
        tsdead_Sfast (float): Proportion of decomposed dead wood that is tranferred to the fast decomposition Soil stock.
        Cgpp (float): Carbon captured in the gross primary productivity of the living forest
        Cr (float):Carbon released by the total respiration of the living forest (i.e. for maintenance and growth)

    Usage:
        >>>grid01=Tree_Grid(x_max=5,y_max=5,delta_h=0.5,patch_area=400,I0=860,lday=12,phi_act=360,k=0.7)
        >>>world01=Tree_World(topology=grid01)
        >>> print(world01)
        World object
    """

    def __init__(self, topology, AET=1300):
        super().__init__(topology)
        self.mortality_factor=1
        self.AET=AET
        self.Smort=0
        self.Sdead={"current":0,"previous":0}
        self.tsdead=min(1.0,np.power(10,(-1.4553+0.0014175*self.AET))/12)
        self.recruit_factor=1.0
        self.dwood_emission_p=0.7
        self.tsdead_A=self.dwood_emission_p*self.tsdead
        self.Sslow={"current":0,"previous":0}
        self.tsdead_Sslow=0.015*0.3*self.tsdead
        self.tsslow_A=1/750
        self.Sfast={"current":0,"previous":0}
        self.tsfast_A=1/15
        self.tsdead_Sfast=0.985*0.3*self.tsdead
        self.smort_log=[]
        self.Cgpp=0
        self.Cr=0
        self.NEE=[]
        self.Stocks={"Sslow":[0,],
                    "Sfast":[0,],
                    "Dwood":[0,],
                    "AGB":[0,],}
        # self.tree_subclasses={"FT1":Tree_FT1,
        #                 "FT2":Tree_FT2,
        #                 "FT3":Tree_FT3,
        #                 "FT4":Tree_FT4,
        #                 "FT5":Tree_FT5,
        #                 "FT6":Tree_FT6}



    def run_simulation(self,n,logging_years,min_dbh,max_dbh,vol,recruitment_month=12,h5file=None,plot_func=None,**kwargs):
        """Run the code in run_schedule() 'n' times.

        Args:
            n (int): Number of simulations to be run.
            logging_years (range): The years in which selective logging occurs.
            min_dbh (float): minimum dbh for a tree to be considered suitable for logging.
            max_dbh (float): maximum dbh for a tree to be considered suitable for logging.
            vol (float): total volume of timber to be extracted by logging event.
            h5file (str): path to the HDF5 file in which model outputs will be saved
            plot_func (function): a function that will be run every step. Usually a plotting function that saves a image file to the disk. Must accept a 'step' argument,which will receive the step number and can be used to generate the file name.
            **kwargs: List of keyword arguments to be passed to plot_func in addition to 'step'.

        Returns:
            None
        """
        if h5file:

            database=open_file(h5file,"r+")

            Pop_table=database.root.sim_1.sys_lvl.Pop
            Stocks_table=database.root.sim_1.sys_lvl.Stocks
            NEE_table=database.root.sim_1.sys_lvl.NEE
            #Emissions_table=database.root.sim_1.sys_lvl.Emissions


            pop_r=Pop_table.row
            stocks_r=Stocks_table.row
            nee_r=NEE_table.row
            #emissions_r=Emissions_table.row


        self.topology.update()
        for t in Tree.Instances.values():
            t.update()

        # if self.step in recruitment_steps:
        #     ext=self.external_seed_rain()
        #     self.incorporate_new_seedbank(ext)
        #     self.stablish_new_trees()

        for i in range(n):

            print("step=",i,"N=",len(Tree.Instances),"DBH>10",Tree.TreesAboveDBH(10.0), "DTrees", sum(Tree.DeadTrees.values()))
            if h5file:
               for ft in Tree.DERIVED_TREES.keys():
                   pop_r[ft]=len(Tree.DERIVED_TREES[ft].Indices)
               pop_r.append()

               stocks_r['agb']=Tree.Total_AGB()*.44
               stocks_r['dead_wood']=self.Sdead["current"]
               stocks_r['soil_slow']=self.Sslow["current"]
               stocks_r['soil_fast']=self.Sfast["current"]
               stocks_r.append()



            self.topology.update()
            for t in Tree.Instances.values():
                t.update()
            self.topology.update()


            if h5file:

               nee_r['dead_wood']=self.tsdead_A*self.Sdead["current"]
               nee_r['soil_slow']=self.tsslow_A*self.Sslow["current"]
               nee_r['soil_fast']=self.tsfast_A*self.Sfast["current"]
               nee_r['gpp']=self.Cgpp
               nee_r['living_trees']=self.Cr
               nee_r['smort']=self.Smort
               nee_r['t_deadwood_sslow']=self.tsdead_Sslow*self.Sdead["previous"]
               nee_r['t_sslow_sfast']=self.tsdead_Sfast*self.Sdead["previous"]



               nee=self.run_schedule(step=i,n=n,h5_database=database)

               nee_r['total_emissions']=-nee+self.Cgpp

               nee_r['nee']=nee
               nee_r.append()
               NEE_table.flush()
               Pop_table.flush()
               Stocks_table.flush()


            else:
                self.NEE.append(self.run_schedule(step=i,n=n))#,h5_database=database)
                self.Stocks["Sslow"].append(self.Sslow["current"])
                self.Stocks["Sfast"].append(self.Sfast["current"])
                self.Stocks["Dwood"].append(self.Sdead["current"])
                self.Stocks["AGB"].append(Tree.Total_AGB()*.44)


            if not plot_func==None: plot_func(step=i,**kwargs)

            if i in logging_years:
                ps=list(self.topology.trees_per_patch.keys())
                st=self.suitable_trees(ps,min_dbh,max_dbh)
                self.log_trees(st,vol)

            # if self.month == recruitment_month:
            #     ext=self.external_seed_rain()
            #     self.incorporate_new_seedbank(ext)
            #     self.stablish_new_trees()


            ext=self.external_seed_rain()
            self.incorporate_new_seedbank(ext)
            self.stablish_new_trees()

            self.increment_time()
            #self.stablish_new_trees()
        if h5file: database.close()



    def run_schedule(self,step,n, h5_database=None):
        """Schedule of actions to be run in every time step

        Args:
            step (int): The current step. Received from run_simulation().
            h5_database (str): path to the HDF5 file in which model outputs will be saved

        Returns:
            (float) the Net Ecosystem Exchange (as calculater by 'calculate_NEE') for thar step.
        """
        #self.topology.update()

        list_of_patches=self.topology.trees_per_patch.keys()
        cohorts=self.topology.cohorts(list_of_patches)
        FTs=list(cohorts.keys())


        for ft in FTs:
            ages=list(cohorts[ft].keys())
            for age in ages:
                base_tree_index=cohorts[ft][age][0]
                base_tree=Tree.Instances.get(base_tree_index)
                base_tree.increase_DBH()
                base_tree.age+=1
                attributes=base_tree.update()

                for tree_index in cohorts[ft][age][1:]:
                    tree=Tree.Instances.get(tree_index)
                    tree.import_attributes(attributes)



        #print("DBH:",Tree.Instances[125].DBH,"AGB:",Tree.Instances[125].AGB )

        #trees=[k for k in Tree.Instances.keys()]
        #for i in trees:
        #    t=Tree.Instances.get(i)
            #print("dbh:",t.id)
        #    t.increase_DBH()

        #

        if h5_database:
            Ind_table=h5_database.root.sim_1.ind_lvl.Ind
            ind_r=Ind_table.row
            trees=[k for k in Tree.Instances.keys()]
            for i in trees:
               t=Tree.Instances.get(i)
            #    #print("up:",t.id)
            #    t.update()
            #    t.age+=1


               ind_r["step"]=step
               ind_r["ind_id"]=t.id
               ind_r["pos_x"]=round(t.position[0],2)
               ind_r["pos_y"]=round(t.position[1],2)
               ind_r["age"]=t.age
               ind_r["dbh"]=t.DBH
               ind_r["agb"]=t.AGB
               ind_r["height"]=t.H
               ind_r["crown_area"]=t.CA
               ind_r["FT"]=t.Ftype
               ind_r.append()

            Ind_table.flush()

        self.Cgpp=Tree.Cgpp()
        self.Cr=Tree.Cr()
        Tree.Mortality()
        #Tree.BackgroundMortality()
        # Tree.CrowdingMortality()
        # Tree.DamageMortality()
        return self.calculate_NEE()


    def create_agents(self,FT,n,pos=None, ids=None, DBHs=None,ages=None, **kwargs):
        """ Creates 'n' trees of type 'FT'. A list of positions, ids and DBHs can be provided (Useful to recreate trees from a previous model run, for example).

        Note: overwrites methods create_agents from class Agent.

        Args:
            agent_type (class): the type (class) of agents to be created
            n (int): number of agents to be created
            pos (list,None): a list of 'n' tuples '(x,y)' with the coordinates in which agents will be created. If set to None, agents are created in random positions.
            ids (list): a list of 'n' ints to be used as tree ids.
            DBHs: a list of 'n' floats to be used as tree DBHs.
            kwargs: the arguments to be passed to the 'agent_type' class. All agents will be created with this same set of arguments.

        Returns:
            None.
        """
        if pos==None:
            pos=[(np.random.rand()*self.topology.x_max,
            np.random.rand()*self.topology.y_max)
            for i in range(n)]

        if ids==None:
            for i in range(n):
                FT(position=pos[i],**kwargs)

        else:
            for i in range(n):
                FT(position=pos[i],dbh=DBHs[i], id=ids[i],age=ages[i],**kwargs)


    def create_HDF_database(self,database_name):



        class NEE(IsDescription):
            dead_wood=Float32Col()
            soil_slow=Float32Col()
            soil_fast=Float32Col()
            living_trees=Float32Col()
            smort=Float32Col()
            t_deadwood_sslow=Float32Col()
            t_sslow_sfast=Float32Col()
            total_emissions=Float32Col()
            gpp=Float32Col()
            nee=Float32Col()


        FTs={ft:Int32Col() for ft in Tree.DERIVED_TREES.keys()}
        Populations=type("Populations",(IsDescription,),FTs)


        class CarbonStocks(IsDescription):
            agb=Float32Col()
            dead_wood=Float32Col()
            soil_slow=Float32Col()
            soil_fast=Float32Col()

        class Individuals(IsDescription):
            step=Int32Col()
            ind_id=Int32Col()
            age=Float32Col()
            pos_x=Float32Col()
            pos_y=Float32Col()
            dbh=Float32Col()
            agb=Float32Col()
            height=Float32Col()
            crown_area=Float32Col()
            FT=StringCol(4)

        h5file=open_file(database_name,mode="w",title="Tree Model outputs")

        sim=h5file.create_group("/","sim_1","Simulation 1")
        sys_lvl=h5file.create_group("/sim_1","sys_lvl","System Level Observations for simulation 1")

        NEE_table=h5file.create_table(sys_lvl,"NEE",NEE,"Net Ecosystem Exchange")
        Pop_table=h5file.create_table(sys_lvl,"Pop",Populations,"Population Sizes")
        Stocks_table=h5file.create_table(sys_lvl,"Stocks",CarbonStocks,"Carbon Stocks")
        #Emissions_table=h5file.create_table(sys_lvl,"Emissions",CarbonEmissions,"Emissions from soil stocks, living trees and dead wood")

        ind_lvl=h5file.create_group("/sim_1","ind_lvl","Individual Level Observations for Simulation 1")

        Ind_table=h5file.create_table(ind_lvl,"Ind",Individuals,"Individual Level Data")

        h5file.close()





    def seedbank_from_file(self, input_file):
        with open(input_file,'r') as json_file:
            self.topology.seedbank=json.load(json_file)



    def seedbank_to_file(self,output_file):
        with open(output_file,'w') as json_file:
            json.dump(self.topology.seedbank,json_file)


    def model_status_from_file(self, input_file):
        with open(input_file,'r') as json_file:
            trees=json.load(json_file)

        for ft in Tree.DERIVED_TREES.keys():
            ages=[]
            positions=[]
            ids=[]
            DBHs=[]
            for tree in trees[ft]:
                positions.append(tuple(tree['position']))
                ids.append(tree['id'])
                DBHs.append(tree['DBH'])
                ages.append(tree['age'])

            self.create_agents(Tree.DERIVED_TREES[ft],len(trees[ft]),pos=positions,DBHs=DBHs,ids=ids, ages=ages, world=self)

        self.topology.update()
        for t in Tree.Instances.values():
            t.update()
            t.AGB=t.AGB_from_DBH(t.DBH/100)
        Tree.ID=max(Tree.Instances.keys())+1


    def model_status_to_file(self,output_file):
        trees_per_type={k:[] for k in Tree.DERIVED_TREES.keys()}
        for t_id,t in Tree.Instances.items():
            trees_per_type[t.Ftype].append({"id":t_id, "position":t.position,"DBH":t.DBH,"age":t.age})

        with open(output_file,'w') as json_file:
            json.dump(trees_per_type,json_file)


    def suitable_trees(self, patches, min_dbh, max_dbh):
        """Search for suitable trees for logging in the given patches.

        Args:
            patches (list): A list of patches (tuples in the format (x,y))
            min_dbh (float): minimum dbh for a tree to be considered suitable for logging.
            max_dbh (float): max dbh for a tree to be considered suitable for logging.

        Returns:
            A dictionary with patches as keys ((x,y)) and lists of suitable tree ids as values
        """

        suitables_per_patch={}
        for p in patches:
            suitables_per_patch[p]=[t for t  in self.topology.trees_per_patch[p] if Tree.Instances[t].DBH>=min_dbh and Tree.Instances[t].DBH<=max_dbh]

        suitables_per_patch={k:v for k,v in suitables_per_patch.items() if v}
        return suitables_per_patch

    def log_trees(self, suitable_trees, total_vol):
        """Randomly log 'suitable_trees' until the 'total_vol' is reached.

        Args:
            suitable_trees (dict): A dictionary with patches as keys ((x,y)) and lists of suitable tree ids as values (resulting from suitable_trees())
            total_vol (float): the total vol of timber to be logged

        Returns:
            None
        """

        total_logged=0
        while total_logged<total_vol and len(suitable_trees):
            ps=list(suitable_trees.keys())
            p=ps[np.random.randint(len(ps))]
            vols={t:Tree.Instances[t].calculate_volume() for t in suitable_trees[p]}
            #pdb.set_trace()
            biggest_tree=max(vols, key=lambda x: vols[x])
            total_logged+=vols[biggest_tree]
            Tree.Instances[biggest_tree].log_me()
            suitable_trees[p].remove(biggest_tree)
            if not suitable_trees[p]:
                del suitable_trees[p]



    def define_seedbank(self,Nseed):
        """Randomly determines how many seed each patch will receive.

        Args:
            Nseed (int):The total number of seed to be distributed among all patches.

        Returns:
            A dictionary in which keys are patches ('(x,y)') and values are the number of seeds
        """

        seedbank={}
        for s in range(Nseed):
            p=(np.random.randint(self.topology.x_max),np.random.randint(self.topology.y_max))
            if seedbank.get(p)==None:
                seedbank[p]=1
            else:
                seedbank[p]+=1

        return seedbank

    def seeds_pos(self,est_seeds):
        """Set a random position within the patch for each seed in est_seeds.

        Args:
            est_seed (dict):A dictionary with patches as keys ('(x,y)') and the number of established seeds as values.

        Returns:
            A list of seed positions within patches.
        """

        pos=[]
        for i in est_seeds.items():
            for j in range(i[1]):
                seed_pos=i[0]+np.random.rand(2)
                pos.append(self.topology.in_bounds(seed_pos))
        return pos



    def external_seed_rain(self):
        external_seedbank={}
        for ft in Tree.DERIVED_TREES.values():
            seedbank_ft=self.define_seedbank(Nseed=int(ft.Nseed*self.topology.total_area))
            pos=self.seeds_pos(seedbank_ft)
            external_seedbank[ft.Ftype]=pos
        return external_seedbank


    def germinate_suitable_seeds(self,seedbank):
        for ft in Tree.DERIVED_TREES.keys():
            pos=self.topology.seedbank[ft]
            self.create_agents(Tree.DERIVED_TREES[ft],len(pos),pos=pos, world=self, dbh=2)

    def incorporate_new_seedbank(self,new_seedbank,clear=True):
        if clear==True:
            self.clear_seedbank()

        for ft in new_seedbank.keys():
            self.topology.seedbank[ft]+=new_seedbank[ft]

    def clear_seedbank(self):
        self.topology.seedbank={k:[] for k in self.topology.seedbank.keys()}

    def stablish_new_trees(self):
        """Create new tree individuals. Use 'define_seedbank()', 'topology.seed_establishment' and 'seeds_pos'.

            Note: This method can be substitued by an equivalent in order to incorporate more realistic representations of seed dispersal.

        Returns:
            None.
        """

        self.topology.seedbank=self.topology.seed_establishment(seedbank=self.topology.seedbank)
        self.germinate_suitable_seeds(seedbank=self.topology.seedbank)
        self.clear_seedbank()



    def calculate_Sdead(self):
        """ Calculate the amount of carbon (tC) in the dead wood stock based
            on the trees that died in the previous year (Smort).

            Returns:
                None.

        """
        #self.Smort/=2
        self.Sdead["current"]=max(0,self.Sdead["previous"]+(self.Smort)-self.tsdead*self.Sdead["previous"])
        self.Smort=0

    def calculate_Sslow(self):
        """ Calculate the amount of carbon (tC) in the slow decomposition soil
            stock.

            Calculation is based on decomposition rates for the dead wood (tsdead_Sslow) and the rate with which carbon is transferred from this stock to the atmosphere (tsslow_A).

            Returns:
                None.
        """

        self.Sslow["current"]=self.Sslow["previous"]+self.tsdead_Sslow*self.Sdead["previous"]-self.tsslow_A*self.Sslow["previous"]

    def calculate_Sfast(self):
        """ Calculate the amount of carbon (tC) in the fast decomposition soil
            stock.

            Calculation is based on decomposition rates for the dead wood (tsdead_Sfast) and the rate with which carbon is transferred from this stock to the atmosphere (tsfast_A).

            Returns:
                None.
        """

        self.Sfast["current"]=self.Sfast["previous"]+self.tsdead_Sfast*self.Sdead["previous"]-self.tsfast_A*self.Sfast["previous"]

    def calculate_NEE(self):
        """
        Return the Net Ecosystem Exchange (NEE) for the previous timestep.

        This is all the carbon that was absorbed by living trees (Cgpp) minus
        what was emitted to the atmosphere by the living trees (Cr) and
        the decomposition of the dead wood and soil stocks (tsdead_A+tsslow_A+tsfast_A).

        Returns:
            None.

        """

        #self.smort_log.append(self.Smort)

        self.calculate_Sdead()
        self.calculate_Sslow()
        self.calculate_Sfast()


        NEE= self.Cgpp-self.Cr-(self.tsdead_A*self.Sdead["previous"])-(self.tsslow_A*self.Sslow["previous"])-(self.tsfast_A*self.Sfast["previous"])

        self.Sdead["previous"]=self.Sdead["current"]
        self.Sslow["previous"]=self.Sslow["current"]
        self.Sfast["previous"]=self.Sfast["current"]



        return NEE



from py_ibm import *
from trees_ibm.tree_agent import Tree
#from plot_trees import plot_tree_pos
import pdb
from tables import *
import json
import numpy as np

class Tree_Grid(Rectangular_Grid):
    """ Represent space as a grid.


    Args:
        x_max (int): number of horizontal cells
        y_max (int): number of vertical cells
        patch_area (float): are of each patch (cells) in squared meters
        I0 (float): Incoming irradiance on top of canopy [umol photon / m^2s]
        k (float): Light extinction coefficient.
        delta_h (float): height of each vertical layer.
        phi_act (float):photosynthetic active period.
        lday (int):mean day length during the vegetation period phi_act.
        ndim (int): number of dimensions. One per environmental variable.
        dim_names (list): a list of 'ndim' strings representing the names of dimensions.

    Attributes:
        LAIs (dict): dictionary with patches as keys and a list of averaged Leaf Area Indexes (one for each vertical layer) as values.
        CCAs (dict):dictionary with patches as keys a a list of Cummulative Crown Areas (one for each vertical layer) as values.
        trees_per_patch (list): dictionary with patches ('(x,y)') as keys and a list of tree ids as values
        total_area(float): area of all simulated patches in ha


    """

    def __init__(self,x_max,y_max,patch_area,I0,k,delta_h,phi_act,lday,ndim=5,dim_names=["GroundLight","LianaCoverage","Temperature","WindSpeed","Humidity"]):
        super().__init__(x_max,y_max,ndim=ndim,dim_names=dim_names)
        self.patch_area=patch_area
        self.delta_h=delta_h
        self.I0=I0
        self.k=k
        self.lday=lday
        self.phi_act=phi_act
        self.LAIs={}
        self.CCAs={}
        self.total_area=self.x_max * self.y_max *self.patch_area/10000
        #s={}
        self.trees_per_patch={(x,y):[] for x in range(0,self.x_max) for y in range(0,self.y_max)}
        #self.FTs=["FT1","FT2","FT3","FT4","FT5","FT6"]
        self.FTs=[k for k in Tree.DERIVED_TREES.keys()]
        self.seedbank={k:[] for k in self.FTs}





    def exeeding_CCA(self):
        patches=[patch for patch,layers in self.CCAs.items() if any(layer>self.patch_area for layer in layers )]

        return patches


    def update(self):
        """ Execute the methods that calculate values for each dimension of the grid.

        Returns:
            None.

        """

        self.LAIs=self.patch_based_LAI(Tree.TreesPerPatch())
        self.CCAs=self.patch_based_CCA(Tree.TreesPerPatch())
        self.ground_level_light()
        self.surface[self.dim_names["LianaCoverage"]]=self.calculate_liana_coverage()

    def seed_establishment(self,seedbank):
        """Check if the seeds in the seedbank have the light and space conditions to germinate.

            Note: All seeds in the given seedbank are assumed to have the same conditions. Multiple calls to this method passing different seedbanks allow the representation of different requirements.

            All established individuals are initiated with a predefined DBH (the same for all individuals), but initial height is calculated based on functional relationships and only if there is enough space in that layer a new individual will be stablished.

        Args:

            seedbank (dict): dictionary with Functional Type (FT) and list of seed positions as values.
            lmin (float): minimum light intensity at ground level for seed to germinate.
            h0 (float): allometric paramenter used to calculate initial height
            h1 (float): allometric paramenter used to calculate initial height

        Returns:
            est_seeds: a dictionary with patches as keys ('(x,y)') and number of established seeds as values.
        """
        est_seeds={}

        for ft in seedbank.keys():
            h0=Tree.DERIVED_TREES[ft].h0
            h1=Tree.DERIVED_TREES[ft].h1
            min_light=Tree.DERIVED_TREES[ft].Iseed
            for i,seed in enumerate(seedbank[ft]):
                patch=(int(np.floor(seed[0])),int(np.floor(seed[1])))
                if not self.check_light(min_light,patch):
                    seedbank[ft].pop(i)
                    break
                if not self.check_CCA(h0,h1,patch):
                    seedbank[ft].pop(i)

        return seedbank



    def check_light(self,min_light, patch):
        """ Check if the light intensity at ground level meets the needs for seed germination.

            Note: Ligth intensity is homogeneized at the patch level.

        Args:
            min_light (float): minimum light intensity at ground for seeds to germinate.
            patch (tuple): (x,y), patch  that will have the light conditions assessed.

        Returns:
            Bool. True if light level is above minimum.
        """
        return self.surface[self.dim_names["GroundLight"]][patch]>min_light

    def check_CCA(self,h0,h1,patch):
        """ Check if the there is enough space forthe new plants to be established

            Note: Cummulative Crown Area is homogeneized at the patch level

        Args:
            h0 (float): allometric paramenter used to calculate initial height
            h1 (float): allometric paramenter used to calculate initial height
            patch (tuple): (x,y), patch  that will have the light conditions assessed.

        Returns:
            Bool. True if there is enough space.
        """
        Hmin=h0*2**h1
        l=int(np.floor(Hmin/self.delta_h))
        if self.CCAs.get(patch)==None:
            return True
        else:
            return self.CCAs[patch][l]<1.0


    def ground_level_light(self):
        """Calculate the amount of light that reaches the ground.

            The value per patch is stored in the "GroundLight" dimension.
            Values are homogeneized at the patch level.

            Returns:
                None.
        """
        self.surface[self.dim_names["GroundLight"]]=self.I0
        for patch in self.LAIs.keys():
            self.surface[self.dim_names["GroundLight"]][patch]=self.I0*np.exp(-self.k*sum(self.LAIs[patch]))



    def patch_based_LAI(self, trees_per_patch):
        """
        Calculate the Leaf Area Index for each patch in the grid.


        Args:
            trees_per_patch (dict): dictionary in which keys are patch coordinates (x,y) and values are  a list of the tree ids in that patch.

        Returns:
            L_per_layer(dict): dictionary in which the keys are patch coordinates('(x,y)')-the same as the ones in trees_per_patch- and values are a list of the LAI per vertical layer
        """

        L_per_layer={}
        nlayers=int(np.ceil(np.ceil(Tree.MaxHeight())/self.delta_h))
        for k in trees_per_patch.keys():
            L_per_layer[k]=[]
            lower_limit=int(min([Tree.Instances.get(t).lmin for t in trees_per_patch[k]]))
            upper_limit=int(max([Tree.Instances.get(t).lmax for t in trees_per_patch[k]]))


            for i in range(nlayers+1):

                L_per_layer[k].append((1/self.patch_area)*sum([Tree.Instances.get(tree_id).L_mean for tree_id in trees_per_patch[k] if lower_limit<=i and upper_limit>=i]))

                #upper and lower limits
                #Tree.Instances.get(tree_id).lmin<=i*self.delta_h and Tree.Instances.get(tree_id).lmax>=i*self.delta_h
        return L_per_layer

    def patch_based_CCA(self, trees_per_patch):
        """Calculate the Leaf Area Index for each patch in the grid.

        Args:
            trees_per_patch (dict): dictionary in which keys are patches ('(x,y)')' and values are lists of the trees ids in that patch.

        Returns:
            L_per_layer(dict): dictionary in which the keys are the same as the ones in trees_per_patch and values are a list of the CCA per vertical layer

            """
        L_per_layer={}
        nlayers=int(np.ceil(np.ceil(Tree.MaxHeight())/self.delta_h))
        for k in trees_per_patch.keys():
            L_per_layer[k]=[]
            lower_limit=int(min([Tree.Instances.get(t).lmin for t in trees_per_patch[k]]))
            upper_limit=int(max([Tree.Instances.get(t).lmax for t in trees_per_patch[k]]))

            for i in range(nlayers+1):

                L_per_layer[k].append((1/self.patch_area)*sum([Tree.Instances.get(tree_id).CA for tree_id in trees_per_patch[k] if lower_limit<=i and upper_limit>=i]))
        return L_per_layer

    def trees_in_patches(self,patches):
        """ Make a list of tree ids and their respective positions.

        Args:
            patch (list): a list of tuples representing the patch coordinates '(x,y)'.

        Returns:
            id_pos (list):a list tuples with tree ids and positions '(id,(x_pos,y_pos))'

        """


        #one list of ids for each patch
        ids=[self.trees_per_patch[p] for p in patches]
        #one single list of tuples with (id, pos)
        id_pos=[(tree,Tree.Instances.get(tree).position) for patch in ids for tree in patch]

        return id_pos


    def trees_per_age_groups(self, list_of_trees):
        """
        Return a dictionary in which keys are a age group and values are lists containing the id of the respective trees.

        list_of_trees is a list with the id of trees.
        """

        age_groups={}
        for t in list_of_trees:
            if age_groups.get(Tree.Instances[t].age)==None:
                age_groups[Tree.Instances[t].age]=[Tree.Instances[t].id]
            else:age_groups[Tree.Instances[t].age].append(Tree.Instances[t].id)

        return age_groups



    def trees_per_type(self, list_of_trees):
        """Organize trees by functional type.

        Args:
            list_of_trees (list): a list of tree ids.

        Returns:
            type_groups(dict): a dictionary in which keys are a Functional
            Types and values are lists containing the id of the respective trees.
        """

        type_groups={}
        for t in list_of_trees:
            if type_groups.get(Tree.Instances[t].Ftype)==None:
                type_groups[Tree.Instances[t].Ftype]=[Tree.Instances[t].id]
            else:type_groups[Tree.Instances[t].Ftype].append(Tree.Instances[t].id)

        return type_groups


    def cohorts(self, patches):
        """ Assemble cohorts based on tree type and age.

            Uses the trees_per_type() and trees_by_age_groups() methods to
            organize a dictionary of dictionaries containing the indices of
            trees in each age group within each type



        Args:
            Patches (list): a list of tuples representing patch coordinatas
            [(x1,y1),(x2,y2)...]

        Returns:
            cohorsts(dict): a dictionary with trees organized by type and age:
             {FunctionalType1:{AgeGroup1:[tree1,tree2,tree3...]}}
        """

        trees_in_patch=self.trees_in_patches(patches)
        list_of_trees=[i[0] for i in trees_in_patch]
        FT=self.trees_per_type(list_of_trees)
        cohorts={}
        for group in FT.keys():
          cohorts[group]=self.trees_per_age_groups(FT[group])

        return cohorts




    def calculate_liana_coverage(self):
        """ Calculate the coverage(%) os lianas in on patch based on the previous coverage, temperature, humidity and wind speed

        Returns:
                None
        """
        return np.zeros((self.x_max,self.y_max))

    def calculate_liana_biomass(self):
        """
        Calculate the biomass of lianas in one patch based on the coverage.
        """
        pass



from setuptools import setup

setup(name='trees_ibm',
      version='0.2.1',
      description='Individual-based tree community models',
      url='http://github.com/fsfrazao/Trees',
      author='Fabio Frazao',
      author_email='fsfrazao@gmail.com',
      license='GNU General Public License v3.0',
      packages=['trees_ibm'],
      install_requires=[
          'numpy',
          'py_ibm',
          ],
      zip_safe=False)
